#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/aruco.hpp>
#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <ctime>

int main(int argc, char* argv[]) {
  if (argc != 9) {
    std::cerr << "Ús: " << argv[0] << " <diccionari> <fitxer_parametres_detector.yml> <files> <columnes> <longitud_marca> <distancia_marques> <fitxer_sortida_calibratge.yml> <camera_id>" << std::endl;
    return -1;
  }

  std::string dictionaryName = argv[1];
  std::string detectorParamsFile = argv[2];
  int rows = std::atoi(argv[3]);
  int cols = std::atoi(argv[4]);
  float markerLength = std::stof(argv[5]);
  float markerSeparation = std::stof(argv[6]);
  std::string outputCalibrationFile = argv[7];

  // Mapa de diccionaris suportats
  std::map<std::string, cv::aruco::PREDEFINED_DICTIONARY_NAME> dictMap = {
    {"DICT_4X4_50", cv::aruco::DICT_4X4_50},
    {"DICT_5X5_100", cv::aruco::DICT_5X5_100},
    {"DICT_6X6_250", cv::aruco::DICT_6X6_250},
    {"DICT_7X7_50", cv::aruco::DICT_7X7_50},
    {"DICT_4X4_100", cv::aruco::DICT_4X4_100}
  };

  if (dictMap.find(dictionaryName) == dictMap.end()) {
    std::cerr << "Error: Diccionari no suportat." << std::endl;
    return -1;
  }

  cv::Ptr<cv::aruco::Dictionary> dictionary = cv::aruco::getPredefinedDictionary(dictMap[dictionaryName]);

  // Carrega de paràmetres del detector
  cv::Ptr<cv::aruco::DetectorParameters> detectorParams = cv::aruco::DetectorParameters::create();
  cv::FileStorage fs(detectorParamsFile, cv::FileStorage::READ);
  if (!fs.isOpened()) {
    std::cerr << "Error obrint fitxer de paràmetres del detector." << std::endl;
    return -1;
  }
  detectorParams->readDetectorParameters(fs.root());
  fs.release();

  cv::Ptr<cv::aruco::GridBoard> board = cv::aruco::GridBoard::create(cols, rows, markerLength, markerSeparation, dictionary);

  cv::VideoCapture inputVideo(std::atoi(argv[8]));
  if (!inputVideo.isOpened()) {
    std::cerr << "No es pot obrir la càmera." << std::endl;
    return -1;
  }

  std::vector<std::vector<cv::Point2f>> allCorners;
  std::vector<int> allIds;
  std::vector<cv::Mat> allImgs;

  int imgCounter = 0;
  cv::Mat frame, frameCopy;

  std::cout << "Prem 'c' per capturar imatge. Prem ESC per acabar." << std::endl;

  while (true) {
    inputVideo >> frame;
    if (frame.empty()) break;

    frame.copyTo(frameCopy);
    std::vector<int> ids;
    std::vector<std::vector<cv::Point2f>> corners;

    cv::aruco::detectMarkers(frame, dictionary, corners, ids, detectorParams);

    // Dibuixar deteccions
    if (ids.size() > 0)
      cv::aruco::drawDetectedMarkers(frameCopy, corners, ids);

    cv::imshow("ArUco Detection", frameCopy);
    char key = (char)cv::waitKey(1);
    if (key == 27) break; // ESC
    if (key == 'c' && ids.size() == (rows*cols)) {
      std::cout << "Capturant imatge " << imgCounter << std::endl;
      std::string filename = "calib_img_" + std::to_string(imgCounter++) + ".jpg";
      cv::imwrite(filename, frame); // guardar imatge sense marques dibuixades
    }
  }
  
  std::cout << "Captura acabada. Iniciant calibratge..." << std::endl;
  
  if (imgCounter < 15) {
    std::cerr << "Mínim 15 imatges requerides per calibrar. Només " << imgCounter << " capturades." << std::endl;
    return -1;
  }

  // Carreguem totes les imatges capturades
  std::vector<std::vector<cv::Point2f>> allCornersConcatenated;
  std::vector<int> allIdsConcatenated;
  std::vector<int> markerCounterPerFrame;

  for (int i = 0; i < imgCounter; ++i) {
    std::string filename = "calib_img_" + std::to_string(i) + ".jpg";
    cv::Mat img = cv::imread(filename);
    if (img.empty()) {
      std::cerr << "No es pot llegir " << filename << std::endl;
      continue;
    }

    std::vector<int> ids;
    std::vector<std::vector<cv::Point2f>> corners;

    cv::aruco::detectMarkers(img, dictionary, corners, ids, detectorParams);

    if (ids.size() > 0) {
      allCornersConcatenated.insert(allCornersConcatenated.end(), corners.begin(), corners.end());
      allIdsConcatenated.insert(allIdsConcatenated.end(), ids.begin(), ids.end());
      markerCounterPerFrame.push_back((int)ids.size());
    }
  }

  if (allIdsConcatenated.empty()) {
    std::cerr << "No s'han detectat suficients marques per calibrar." << std::endl;
    return -1;
  }

  cv::Mat cameraMatrix, distCoeffs;
  std::vector<cv::Mat> rvecs, tvecs;

  double repError = cv::aruco::calibrateCameraAruco(
    allCornersConcatenated,
    allIdsConcatenated,
    markerCounterPerFrame,
    board,
    cv::Size(frame.cols, frame.rows),
    cameraMatrix,
    distCoeffs,
    rvecs,
    tvecs
  );

  std::cout << "Error de reprojecció mitjà: " << repError << std::endl;

  // Desa els resultats al fitxer YAML
  cv::FileStorage fsCalib(outputCalibrationFile, cv::FileStorage::WRITE);
  if (!fsCalib.isOpened()) {
    std::cerr << "No es pot escriure el fitxer de calibratge: " << outputCalibrationFile << std::endl;
    return -1;
  }

  fsCalib << "camera_matrix" << cameraMatrix;
  fsCalib << "distortion_coefficients" << distCoeffs;
  fsCalib << "reprojection_error" << repError;

  fsCalib.release();
  std::cout << "Calibratge complet. Resultats desats a: " << outputCalibrationFile << std::endl;

  return 0;
}

