#include "opencv2/core/core.hpp"
#include "opencv2/aruco.hpp"
#include "opencv2/highgui.hpp"
#include <iostream>
#include <string>
#include <map>
#include <vector>

int main(int argc, char* argv[]) {
    // Comprovem si hi ha prou paràmetres en la línia de comandes
    if (argc != 7) {
        std::cerr << "Usage: " << argv[0] << " <rows> <cols> <dictionary> <marker_size> <separation> <output_file>" << std::endl;
        return -1;
    }

    // Paràmetres de la línia de comandes
    int rows = std::atoi(argv[1]);             // Nombre de files
    int cols = std::atoi(argv[2]);             // Nombre de columnes
    std::string dictionaryName = argv[3];      // Nom del diccionari (ex. DICT_4X4_50)
    int markerSize = std::atoi(argv[4]);      // Mida de cada marca
    int separation = std::atoi(argv[5]);      // Separació entre marques
    std::string outputFile = argv[6];         // Nom del fitxer PNG de sortida

    // Mapa de diccionaris possibles
    std::map<std::string, cv::Ptr<cv::aruco::Dictionary>> dictionaryMap = {
        {"DICT_4X4_50", cv::aruco::getPredefinedDictionary(cv::aruco::DICT_4X4_50)},
        {"DICT_5X5_100", cv::aruco::getPredefinedDictionary(cv::aruco::DICT_5X5_100)},
        {"DICT_6X6_250", cv::aruco::getPredefinedDictionary(cv::aruco::DICT_6X6_250)},
        {"DICT_7X7_50", cv::aruco::getPredefinedDictionary(cv::aruco::DICT_7X7_50)},
        // Afegir més diccionaris segons les necessitats
    };

    // Comprovem si el diccionari passat existeix en el mapa
    if (dictionaryMap.find(dictionaryName) == dictionaryMap.end()) {
        std::cerr << "Error: Diccionari no suportat o incorrecte!" << std::endl;
        return -1;
    }

    // Obtenir el diccionari seleccionat
    cv::Ptr<cv::aruco::Dictionary> dictionary = dictionaryMap[dictionaryName];

    // Crear una matriu per a la imatge final sense borde
    int totalWidth = cols * (markerSize + separation) - separation;
    int totalHeight = rows * (markerSize + separation) - separation;

    // Afegir un borde de separation píxels a tota la imatge
    cv::Mat boardImageWithBorder(totalHeight + 2 * separation, totalWidth + 2 * separation, CV_8UC1, cv::Scalar(255)); // Fons blanc

    // Crear una matriu per a la imatge de la matriu ArUco
    cv::Mat boardImage(totalHeight, totalWidth, CV_8UC1, cv::Scalar(255)); // Fons blanc

    // Afegir les marques a la matriu
    int markerId = 0; // Iniciem l'ID de la marca
    for (int row = 0; row < rows; ++row) {
        for (int col = 0; col < cols; ++col) {
            // Crear la imatge del marker
            cv::Mat markerImage;
            cv::aruco::drawMarker(dictionary, markerId, markerSize, markerImage);

            // Calcular la posició de la marca a la matriu
            int xOffset = col * (markerSize + separation);
            int yOffset = row * (markerSize + separation);

            // Copiar la imatge del marker a la matriu sense borde
            markerImage.copyTo(boardImage(cv::Rect(xOffset, yOffset, markerSize, markerSize)));

            // Incrementar l'ID de la marca
            markerId++;

            // Si hem arribat a la quantitat de marques necessàries (files x columnes), parem
            if (markerId >= rows * cols) {
                break;
            }
        }
        if (markerId >= rows * cols) {
            break;
        }
    }

    // Copiar la matriu ArUco amb les marques dins la imatge amb borde
    boardImage.copyTo(boardImageWithBorder(cv::Rect(separation, separation, totalWidth, totalHeight)));

    // Desar la imatge amb borde en format PNG
    if (!cv::imwrite(outputFile, boardImageWithBorder)) {
        std::cerr << "Error: No s'ha pogut desar la imatge!" << std::endl;
        return -1;
    }

    std::cout << "Matriu ArUco generada amb borde i desada a: " << outputFile << std::endl;

    return 0;
}

